/******************************************************************************************

   Fiver.java - Fiver puzzle
   Version 2.0
   www.mazeworks.com
   
   Copyright (c) 2002 Robert Kirkland
   All Rights Reserved.

   Permission is hereby granted, free of charge, to any person obtaining a copy 
   of this software and associated documentation files (the "Software"), to
   deal in the Software without restriction, including without limitation the
   rights to use, copy, modify, merge, publish, distribute, and/or sell copies
   of the Software, and to permit persons to whom the Software is furnished to
   do so, provided that the above copyright notice and this permission notice
   appear in all copies of the Software and that both the above copyright
   notice and this permission notice appear in supporting documentation.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS.
   IN NO EVENT SHALL THE COPYRIGHT HOLDER INCLUDED IN THIS NOTICE BE LIABLE FOR
   ANY CLAIM, OR ANY SPECIAL INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Except as contained in this notice, the name of the copyright holder shall
   not be used in advertising or otherwise to promote the sale, use or other
   dealings in this Software without prior written authorization of the
   copyright holder.

******************************************************************************************/

import java.applet.* ;
import java.awt.* ;
import java.awt.image.*;
import java.net.* ;
import java.util.* ;
	
/******** FIVER ********
	main class: initializes applet panel, 
	starts games, handles control panel & moves
*/	
public class Fiver extends Applet {
   // applet info
   public String getAppletInfo() { return "Fiver 2.0 - Copyright (c) 2002 Robert Kirkland"; }
   String parameterInfo[][] = {
      {"bgColor", "integer", "Background color (24-bit RGB hex int)"},
      {"title",   "string",  "Applet title text"},
      {"restart", "string",  "Restart button text"},
      {"board",   "string",  "Board label text"},
      {"bd3",     "string",  "3x3 Board size text"},
      {"bd4",     "string",  "4x4 Board size text"},
      {"bd5",     "string",  "5x5 Board size text"},
      {"bd6",     "string",  "6x6 Board size text"},
      {"bd7",     "string",  "7x7 Board size text"},
      {"bd8",     "string",  "8x8 Board size text"},
      {"moves",   "string",  "Moves label text"},
      {"win",     "string",  "Win status text"},
      {"perfect", "string",  "Pefect game status text"} } ;
   public String[][] getParameterInfo() { return parameterInfo ; }

   static final Color  DF_BG_COLOR  =Color.white ;   // default background color
   static final String DF_TITLE  ="FIVER",           // default text strings
                       DF_RESTART="RESTART",
                       DF_BOARD  ="BOARD",
                       DF_BD3    =" 3 x 3",
                       DF_BD4    =" 4 x 4",
                       DF_BD5    =" 5 x 5",
                       DF_BD6    =" 6 x 6",
                       DF_BD7    =" 7 x 7",
                       DF_BD8    =" 8 x 8",
                       DF_MOVES  ="MOVES",
                       DF_WIN    =" You win !",
                       DF_PERFECT=" Perfect !" ;
   static final Font titleFont =new Font("SansSerif", Font.BOLD, 14),
                     textFont  =new Font("SansSerif", Font.PLAIN, 12),
                     boardFont =new Font("Courier", Font.PLAIN, 12),
                     movesFont =new Font("Courier", Font.BOLD, 14),
                     statusFont=new Font("Dialog", Font.PLAIN, 12) ;
   static final int BOARD_WIDTH=240, BOARD_HEIGHT=240 ;
   protected Color BG_COLOR ;
   protected String TITLE, RESTART, BOARD, BD3, BD4, BD5, BD6, BD7, BD8,
                    MOVES, WIN, PERFECT ;
   static final int bcWidth=360, bcHeight=360,
                    sizeMoves[][] = { {3,5},{4,4},{5,15},{6,28},{7,33},{8,40} } ;
   private int moves=0 ;
   private boolean gameOver ;
   private Board bd ;
   private BoardCanvas bc ;
   private Panel cp ;
   private Button bRestart ;
   private Choice cBoardSizes ;
   private TextField tfMoveCount, tfStatus ;
   private Label title ;
   private Image gifletsImage ;

   public void init() {
      // get applet parameters
      // background color
      String bgColorString = getParameter("bgColor") ;
      try { BG_COLOR = bgColorString==null ? DF_BG_COLOR : new Color(Integer.parseInt(bgColorString,16)) ; } 
      catch (NumberFormatException e) {
         BG_COLOR = DF_BG_COLOR ;
         System.out.println("ERROR: Invalid bgColor parameter.") ;
      }
      // text
      TITLE = getParameter("title") ; if (TITLE==null) TITLE = new String(DF_TITLE) ;
      RESTART = getParameter("restart") ; if (RESTART==null) RESTART = new String(DF_RESTART) ;
      BOARD = getParameter("board") ; if (BOARD==null) BOARD = new String(DF_BOARD) ;
      BD3 = getParameter("bd3") ; if (BD3==null) BD3 = new String(DF_BD3) ;
      BD4 = getParameter("bd4") ; if (BD4==null) BD4 = new String(DF_BD4) ;
      BD5 = getParameter("bd5") ; if (BD5==null) BD5 = new String(DF_BD5) ;
      BD6 = getParameter("bd6") ; if (BD6==null) BD6 = new String(DF_BD6) ;
      BD7 = getParameter("bd7") ; if (BD7==null) BD7 = new String(DF_BD7) ;
      BD8 = getParameter("bd8") ; if (BD8==null) BD8 = new String(DF_BD8) ;
      MOVES = getParameter("moves") ; if (MOVES==null) MOVES = new String(DF_MOVES) ;
      WIN = getParameter("win") ; if (WIN==null) WIN = new String(DF_WIN) ;
      PERFECT = getParameter("perfect") ; if (PERFECT==null) PERFECT = new String(DF_PERFECT) ;

      // load images
      MediaTracker tracker = new MediaTracker(this) ;
      URL url = getCodeBase() ;
      gifletsImage = getImage(url,"giflets.gif") ;
      tracker.addImage(gifletsImage,0) ;
      try { tracker.waitForAll(); }
      catch (InterruptedException e) {}
      // applet layout 
      setBackground(BG_COLOR) ;
      setLayout(new FlowLayout(FlowLayout.CENTER,10,10)) ;
      bc = new BoardCanvas(bcWidth,bcHeight,gifletsImage,this) ;
      bc.resize(bcWidth,bcHeight) ;

      // create Control Panel

      cp = new Panel() ;
      cp.setLayout(new GridLayout(0,1,0,5)) ;
      cp.setBackground(Color.lightGray) ;
      cp.setFont(textFont) ;
      // Title
      cp.add(title=new Label(TITLE,Label.CENTER)) ;
      title.setFont(titleFont) ;
      cp.add(new Label(" ")) ;
      // Restart button
      cp.add(bRestart=new Button(RESTART)) ;
      // Board pulldown
      cp.add(new Label(BOARD,Label.CENTER)) ;
      cp.add(cBoardSizes=new Choice()) ;
      cBoardSizes.setFont(boardFont) ;
      cBoardSizes.setForeground(Color.black) ;
      cBoardSizes.setBackground(Color.white) ;
      cBoardSizes.addItem(BD3);
      cBoardSizes.addItem(BD4);
      cBoardSizes.addItem(BD5);
      cBoardSizes.addItem(BD6);
      cBoardSizes.addItem(BD7);
      cBoardSizes.addItem(BD8);
      cBoardSizes.select(BD5) ;
      // MoveCount textbox
      cp.add(new Label(MOVES,Label.CENTER)) ;
      cp.add(tfMoveCount=new TextField(6)) ;      
      tfMoveCount.setFont(movesFont) ;
      tfMoveCount.setForeground(Color.black) ;
      tfMoveCount.setBackground(Color.white) ;
      tfMoveCount.setEditable(false) ;
      // Status textbox
      cp.add(tfStatus=new TextField(12)) ;
      tfStatus.setFont(statusFont) ;
      tfStatus.setForeground(Color.white) ;
      tfStatus.setBackground(Color.black) ;
      tfStatus.setEditable(false) ;
      // create enclosing panels
      Panel p1 = new Panel() ;
      p1.setBackground(Color.lightGray) ;
      p1.setLayout(new FlowLayout(FlowLayout.CENTER,5,5)) ;
      Panel p2 = new Panel() ;
      p2.setBackground(Color.gray) ;
      p2.setLayout(new FlowLayout(FlowLayout.CENTER,2,2)) ;
      p1.add(cp) ;
      p2.add(p1) ;
      add(bc) ;
      add(p2) ;
      validate() ;

		newGame() ;
	}
   // start a new game
	void newGame() {
		System.gc() ;
      gameOver = false ; 
      moves = 0 ;
      setMoveCount(moves) ;
      tfMoveCount.setForeground(Color.black) ;
      setStatus(" ") ;
      bd = new Board(this) ;
      bc.drawBoard(bd) ;
      bc.requestFocus() ;
   }      
   private void endGame() {
      if (moves==sizeMoves[getBoardSelected()][1])
         setStatus(PERFECT) ;
      else setStatus(WIN) ;
      gameOver = true ;
   }
	// handle board mouse event
   void playerMove(int x, int y) {

      if (!gameOver) {
         if (bd.move(x,y)) {
            bc.drawBoard(bd) ;
            setMoveCount(++moves) ;
            if (bd.isWin()) endGame() ;
         }
      }
   }
   // handle control panel event
   public boolean action(Event e,Object o) { 
      newGame() ;
      return true ;
   }
   // update Moves field
   void setMoveCount(int i) { 
      String s = Integer.toString(i) ;
      switch (s.length()) {
         case 1: s = "      " + s ; break ;
         case 2: s = "     " + s ; break ;
         case 3: s = "    " + s ; break ;
         case 4: s = "   " + s ; break ;
         case 5: s = "  " + s ; break ;
         case 6: s = " " + s ; break ;
      }
      if (i==sizeMoves[getBoardSelected()][1]) tfMoveCount.setForeground(Color.red) ;
      else if (i>sizeMoves[getBoardSelected()][1]) tfMoveCount.setForeground(Color.gray) ;
      tfMoveCount.setText(s) ;
   }
   void setStatus(String s) { tfStatus.setText(s!=null?s:"") ; }
   int getBoardSelected() { return cBoardSizes.getSelectedIndex() ; }
   int getDimension() { return sizeMoves[getBoardSelected()][0] ; }
}      
/******** BOARD ********
	initializes board, makes moves, checks for win
*/	
final class Board {
   // all boards are a 64-bit string, where black=0, white=1
   static final long startingBoards[] = 
                       { 0x1FFL, 0xFFFFL, 0x1FFFFFFL, 0xFFFFFFFFFL,
                         0x1FFFFFFFFFFFFL, 0xFFFFFFFFFFFFFFFFL } ;
   private int square, size ;
   private long board ;
   private Fiver main ;
	
   // constructor
   Board(Fiver main) {
      this.main = main ;
      // initialize board
      board = startingBoards[main.getBoardSelected()] ;
      size = main.getDimension() ;
   }
	boolean isValidSquare(int x,int y) { 
      if ((x>=0)&&(x<size)&&
          (y>=0)&&(y<size)) return true ;
      else return false ;
   }
   boolean move(int x,int y) {
      if (isValidSquare(x,y)) {
         // convert x,y to 1-dimension
         square = (y * size) + x ;
         // flip selected square
         board ^= (1L<<square) ;
         // flip north
         if (y>0) board ^= (1L<<(square-size)) ;
         // flip south
         if (y<(size-1)) board ^= (1L<<(square+size)) ;
         // flip east
         if (x<(size-1)) board ^= (1L<<(square+1)) ;
         // flip west
         if (x>0) board ^= (1L<<(square-1)) ;
         return true ;
      }
      else return false ;
   }
   boolean isWin() {
      if (board==0) return true ;
      else return false ;
   }
   long getBoard() { return board ; }
   int getBoardSize() { return size ; }
}			
/******** BOARD CANVAS ********
	controls all graphics, handles all mouse events 
*/	
final class BoardCanvas extends Canvas {
   // giflet coordinates
   final static int lightSquareCoord[]={0,0,42,42}, darkSquareCoord[]={43,0,42,42},
      whitePieceCoord[]={0,42,34,34},blackPieceCoord[]={36,42,34,34},
      startOffset[] = {110,90,70,50,30,10}, SQUARE=41, OFFSET=5 ;
   final static Color frameColor=new Color(204,153,0),
                      frameHighColor=new Color(255,204,102) ;
   private int canvasWidth, canvasHeight, bdSelected ;
	private Image bufferImage, gifletsImage, lightSquare, darkSquare, 
                 whitePiece, blackPiece ;
   private MediaTracker tracker ;
	private Graphics buffer ;
	private Fiver main ;

	BoardCanvas(int canvasWidth,int canvasHeight,Image gifletsImage,Fiver main) { 
      this.canvasWidth = canvasWidth ;
      this.canvasHeight = canvasHeight ;
      this.gifletsImage = gifletsImage ;
      this.main = main ;

      // Extract little images from giflets GIF
      tracker = new MediaTracker(main) ;
      lightSquare = extractImage(lightSquareCoord) ;
      darkSquare = extractImage(darkSquareCoord) ;
      whitePiece = extractImage(whitePieceCoord) ;
      blackPiece = extractImage(blackPieceCoord) ;
      try { tracker.waitForAll() ; }
      catch (InterruptedException e) {}
   }
	void drawBoard(Board bd) {
      boolean light ;
      int row, column, x, y, bdSize=bd.getBoardSize() ;
      long board = bd.getBoard() ;

		if (buffer==null) {
			bufferImage = createImage(canvasWidth,canvasHeight) ;
			buffer = bufferImage.getGraphics() ;
		}
      buffer.setColor(main.BG_COLOR) ;
      buffer.fillRect(0,0,canvasWidth,canvasHeight) ;
		// draw the board and pieces
      bdSelected = main.getBoardSelected() ;
      row = column = 0 ;
      light = true ;
      for (int i=0; i<bdSize; i++) {
         if (i%2==0) light = true ;
         else light = false ;
         for (int j=0; j<bdSize; j++) {
            x = startOffset[bdSelected] + (column*SQUARE) ;
            y = startOffset[bdSelected] + (row*SQUARE) ;
            // draw single square
            if (light) {
               buffer.drawImage(lightSquare,x,y,this) ;
               light = false ;
            }
            else {
               buffer.drawImage(darkSquare,x,y,this) ;
               light = true ;
            }
            // draw piece for square
            if ( (board&(1L<<((row*bdSize)+column)))==0 )
               buffer.drawImage(blackPiece,x+3,y+3,this) ;
            else buffer.drawImage(whitePiece,x+3,y+3,this) ;
            column++ ;
         }
         column = 0 ;
         row++ ;
      }
      // draw frame
      x = (bdSize * SQUARE) + 2 ;
      buffer.setColor(frameColor) ;
      buffer.drawRect(startOffset[bdSelected]-1,startOffset[bdSelected]-1,x,x) ;
      buffer.drawRect(startOffset[bdSelected]-3,startOffset[bdSelected]-3,x+4,x+4) ;
      buffer.setColor(frameHighColor) ;
      buffer.drawRect(startOffset[bdSelected]-2,startOffset[bdSelected]-2,x+2,x+2) ;

      repaint() ;
   }
	public void paint(Graphics g) { update(g) ; }
	public void update(Graphics g) { g.drawImage(bufferImage,0,0,this) ; }
	
	// board canvas mouse event handler
   public boolean mouseDown(Event e,int x,int y) {   
      if (x<360) { 
         int boardX = x - startOffset[bdSelected] ;
         int boardY = y - startOffset[bdSelected] ;
         if ( (boardX%SQUARE>OFFSET) && (boardY%SQUARE>OFFSET) &&
              (boardX%SQUARE<(SQUARE-OFFSET)) && (boardY%SQUARE<(SQUARE-OFFSET)) )
            main.playerMove(boardX/SQUARE,boardY/SQUARE) ;
      }
      return true ; 
   }
   // Extracts GIF images
   private Image extractImage(int[] xyCoord) {
      Image newImage ;
      ImageFilter filter;
      ImageProducer producer;

      filter = new CropImageFilter(xyCoord[0],xyCoord[1],xyCoord[2],xyCoord[3]);
      producer = new FilteredImageSource(gifletsImage.getSource(), filter);
      newImage = main.createImage(producer);
      tracker.addImage(newImage,0);
      return newImage ;
   }
}